//(c) Artem Parlyuk, PhD., 
// e-mail: info@artsoft.name, http://artsoft.name
//(c) Boris Y. Lemeshko, Prof., 
// Department of Applied Mathematics and Computer Science
// Novosibirsk State Technical University, 
// e-mail: heard@fpm.ami.nstu.ru


#include "minimize.hpp"

/* p  */
//    p p
// -p- -1
// p                 -2
// p                 -3

 void (*grad)(double (*f)(double*), double*, double*, int) = Grad;

 void SetGrad(void (*fun)(double (*f)(double*), double*, double*, int))
 {
   grad = fun;
 }

 double metrik(double (*f)(double x[]),double x[],int n,
	       void (*bldflp)(int),double eps1,double eps2);

 void blisk(double x[],double xx[],int n,double a,double sag[],double *da,
		  int *is,double (*f)(double x[]));
 void blbrdn(int);
 void bldflp(int);
 void blflet(int);
 double blnrm(double x[],int n);

 double QuasiNewton(double (*f)(double x[]),double x[],int n,
	       int param, double eps1, double eps2)
 {
  switch(param)
  {
   case 2: return metrik(f, x, n, blflet, eps1, eps2);
   case 3: return metrik(f, x, n, blbrdn, eps1, eps2);
   case 1:
   default: return metrik(f, x, n, bldflp, eps1, eps2);
  }
 }	       
 

/* double h[10][10], h1[10][10], h2[10][10], y[10], yy1[10], sm[10],
	  sag[10], df[10], dx[10], x1[10]*/;
	  
double **h,**h1,**h2;
double *y,*yy1,*sm,*sag,*df,*dx,*x1;

double metrik(double (*f)(double x[]),double x[],int n,
	       void (*bldflp)(int),double eps1,double eps2)
{
 double  da, s, a1, a;
 int  i, itw, j, is, neud, ks;

 y = new double[n];
 yy1 = new double[n];
 sm = new double[n];
 sag = new double[n];
 df = new double[n];
 dx = new double[n];
 x1 = new double[n];
 h = new double*[n];
 h1 = new double*[n];
 h2 = new double*[n];
 for(int i=0;i<n;i++)
 {
  h[i] = new double[n];
  h1[i] = new double[n];
  h2[i] = new double[n];
 }

   iteration = 0;
   minerror = 0;

  for(i=0; i<n; i++)
	x1[i]=x[i];

 a=(*f)(x);

 if(eps1==0.)
	{ eps1=0.1e-3; eps2=eps1;}

L4: 
          if(Minimize_callback) Minimize_callback(a, x, n);
	  if((maxiteration > 0) && (iteration > maxiteration))
	  {
	   minerror = ERR_ITER;
	   goto L50;
	  }
        if(minimize_stop)
	{
	 minerror = ERR_ITER;
	 goto L50;
	}
          iteration ++;
  da=0.1;
  itw=0;
  for(i=0; i<n; i++)
	for(j=0; j<n; j++)
	  if(i==j)
		 h[i][j]=1.;
	  else
		 h[i][j]=0.;

  (*grad)(f, x, df, n);

L9:
 s=sqrt(blnrm(df,n));

 if(s==0.) goto L50;

  for(i=0; i<n; i++)
    {
	 sag[i]=0.;
	 for(j=0; j<n; j++)
	    sag[i]=sag[i]-h[i][j]*df[j];
    }
  blisk(x,yy1,n,a,sag,&da,&is,(*f));
 if(is==1) goto L60;

 neud=0;
 for(i=0; i<n; i++)
 {
   sm[i]=da*sag[i];
   y[i]=-df[i];
   x[i]=x[i]+sm[i];
  }

 a1=a;
 a=(*f)(x);
 ks=0;
 itw++;
 if( fabs(a1) > 1. ) 
	{if( fabs((a1-a)/a1) < eps2 ) goto L50;}
 else	
	{if( fabs(a1-a) < eps2 ) goto L50;}

  for(i=0; i<n; i++)
   {
    if(fabs(x[i]) > 1.) 
	   {if( fabs(sm[i]/x[i]) < eps1) ks++;}
    else
	   if(fabs(sm[i]) < eps1) ks++;
   }
 if(ks==n) goto L50;

  if(itw>2*n+1) goto L4;
  (*grad)(f, x, df, n);

  for(i=0; i<n; i++)
	 y[i]=y[i]+df[i];

  (*bldflp)(n);
  goto L9;

L60:
  for(i=0; i<n; i++)
    {
	if(x1[i]!=x[i])
	 {
	  for(j=0; j<n; j++)
		  x1[j]=x[j];
	  if(neud==1) goto L50;
	  neud=1;
	  goto L4;
	 }
    }

L50:

 delete y;
 delete yy1;
 delete sm;
 delete sag;
 delete df;
 delete dx;
 delete x1;
 for(int i=0;i<n;i++)
 {
  delete h[i];
  delete h1[i];
  delete h2[i];
 }
 delete h;
 delete h1;
 delete h2;
 return a;
}

/*  p p

 double blnrm(double x[],int n)
 {
 double s;
 int i;

 s=0.;
  for(i=0; i<n; i++)
	 s+=x[i]*x[i];
 return(s);
}

 void blisk(double x[],double xx[],int n,double a,double sag[],double *da,
		  int *is,double (*f)(double x[]))
pp  p  
 n p  p p sag
    x(n) -   pp ;
    sag(n) - p ;
    a -     x(n);
    da -  , ppp  pp;
    f - p p ;
    xx(n) - p ;
    mot - ,   p .

{
 double  r, t, s;
 int  i, k;

 *is=0;

 for(i=0; i<n; i++)
   if(sag[i]!=0.) goto L32;
 *is=1;
 return;

L32:
 k=0;
 for(i=0; i<n; i++)
	 xx[i]=x[i]+*da*sag[i];
 r=(*f)(xx);
 if(a<r) goto L5;

L2:
  for(i=0; i<n; i++)
	 xx[i]=xx[i]+*da*sag[i];
 t=(*f)(xx);
 if(r<t) goto L10;
 *da*=2;
 r=t;
 goto L2;

L5:
 k++;
 *da/=2;
 t=r;
 for(i=0; i<n; i++)
	 xx[i]=x[i]+*da*sag[i];
 r=(*f)(xx);
 if( r <= a ) goto L10;
 if( k <= 20 ) goto L5;
 *is=1;
 return;

L10:
 s=a+t-2*r;
 if( s == 0. ) return;

L11:
 s=*da*(a*3+t-4*r)/(s*2);
 *da=s;
 return;

}
*/
void blbrdn(int n)
{
 double p;
 int i, j;

  for(i=0; i<n; i++)
	 yy1[i]=0.;
  for(i=0; i<n; i++)
  {
	 for(j=0; j<n; j++)
		 yy1[i]+=h[i][j]*y[j];
	 yy1[i]=sm[i]-yy1[i];
  }
 p=0.;
  for(i=0; i<n; i++)
	 p+=yy1[i]*y[i];
  for(i=0; i<n; i++)
	  for(j=0; j<n; j++)
		 h[i][j]+=yy1[i]*yy1[j]/p;
}

 void bldflp(int n)
{
 double p, p1;
 int i, j;

 p=0.;
  for(i=0; i<n; i++)
   {
	p+=sm[i]*y[i];
	yy1[i]=0.;
   }
 if(p==0.) return;
 for(i=0; i<n; i++)
    for(j=0; j<n; j++)
	   yy1[i]+=h[i][j]*y[j];
 p1=0.;
  for(i=0; i<n; i++)
	 p1+=y[i]*yy1[i];
  for(i=0; i<n; i++)
  for(j=0; j<n; j++)
		 h[i][j]+=sm[i]*sm[j]/p-yy1[i]*yy1[j]/p1;
}

 void blflet(int n)
{
 double p;
 int i, j, k;

 p=0.;
 for(i=0; i<n; i++)
	p+=sm[i]*y[i];

 for(i=0; i<n; i++)
   for(j=0; j<n; j++)
   {
	h2[i][j]=0.;
	if(i==j)
		h1[i][j]=1.-y[i]*sm[j]/p;
	else
		h1[i][j]=-y[i]*sm[j]/p;
    }

 for(i=0; i<n; i++)
    for(j=0; j<n; j++)
	  for(k=0; k<n; k++)
			 h2[i][j]+=h[i][k]*h1[k][j];

  for(i=0; i<n; i++)
	for(j=0; j<n; j++)
	   h[i][j]=0.;

  for(i=0; i<n; i++)
	for(j=0; j<n; j++)
	  {
	   for(k=0; k<n; k++)
		 h[i][j]+=h1[k][i]*h2[k][j];
	   h[i][j]+=y[i]*sm[j]/p;
	  }
}

void Grad(double (*f)(double*), double* x, double* dx, int n)
{
 double h = 1e-8, h2 = 2*h, *x1, f1, f0;
 int i;
 x1 = new double[n];
  for(i = 0; i < n; i++) x1[i] = x[i];
  for(i = 0; i < n; i++) {
   x1[i] = x[i] - h; f0 = f(x1);
   x1[i] = x[i] + h; f1 = f(x1);
   x1[i] = x[i];
   dx[i] = (f1 - f0) / h2; }
 delete x1;
}

double Davidon(double (*f)(double x[]),double x[],int n,
	       double eps1, double eps2)
{
 return metrik(f, x, n, bldflp, eps1, eps2);
}	       

// p                 

double Fletcher(double (*f)(double x[]),double x[],int n,
	       double eps1, double eps2)
{
 return metrik(f, x, n, blflet, eps1, eps2);
}	       

// p                 

double Broyden(double (*f)(double x[]),double x[],int n,
	       double eps1, double eps2)
{
 return metrik(f, x, n, blbrdn, eps1, eps2);
}	       

