//(c) Artem Parlyuk, PhD., 
// e-mail: info@artsoft.name, http://artsoft.name
//(c) Boris Y. Lemeshko, Prof., 
// Department of Applied Mathematics and Computer Science
// Novosibirsk State Technical University, 
// e-mail: heard@fpm.ami.nstu.ru

#include "minimize.hpp"

/* p  */

 void blisk(double x[],double xx[],int n,double a,double sag[],double *da,
		  int *is,double (*f)(double x[]));
 double blnrm(double x[],int n);
 void inv(double** a,int n,int *iflag);
 void gord(double** a,int n,int k,int l);

void (*hesse)(double (*f)(double*), double*, double** dx, int) = Hesse;

void SetHesse(void (*fun)(double (*f)(double*), double*, double**, int))
{
 hesse = fun;
}


extern double** h;
extern double *sm, *yy1, *sag, *df, *dx, *x1;
int* m;


 /*    H */

 double Newton(double (*f)(double x[]), 
              double x[],int n,double eps1, double eps2)
{
 double  da, s, a1, a;
 int  i, itw, j, is, neud, ks, iflag;

 sm = new double[n];
 yy1 = new double[n];
 sag = new double[n];
 df = new double[n];
 dx = new double[n];
 x1 = new double[n];
 h = new double*[n];
 m = new int[6*n];
 
 for(i=0;i<n;i++) h[i] = new double[n];
 
   iteration = 0;
   minerror = 0;

  for(i=0; i<n; i++)
	x1[i]=x[i];

 a=(*f)(x);

 if(eps1==0.)
	{ eps1=0.1e-3; eps2=eps1;}

L4:
        if(Minimize_callback) Minimize_callback(a, x ,n);
	if((maxiteration > 0) && (iteration > maxiteration)) 
	{
	 minerror = ERR_ITER;
	 goto L50;
	}
        if(minimize_stop)
	{
	 minerror = ERR_ITER;
	 goto L50;
	}
        iteration++;

  da=0.1;
  itw=0;
  (*hesse)(f,x,h,n);
  inv(h,n,&iflag);
  if(!iflag) goto L50;
  (*grad)(f,x,df,n);

L9:
 s=sqrt(blnrm(df,n));

 if(s==0.) goto L50;

 for(i=0; i<n; i++)
    {
	 sag[i]=0.;
	 for(j=0; j<n; j++)
	    sag[i]=sag[i]-h[i][j]*df[j];
    }

 blisk(x,yy1,n,a,sag,&da,&is,(*f));
 if(is==1) goto L60;

/*                */
/*    .       */
/*       */
/* da=1; */

 neud=0;
 for(i=0; i<n; i++)
 {
   sm[i]=da*sag[i];
   x[i]=x[i]+sm[i];
  }

 a1=a;
 a=(*f)(x);
 ks=0;
 itw++;
 if( fabs(a1) > 1. ) 
	{if( fabs((a1-a)/a1) < eps2 ) goto L50;}
 else	
	{if( fabs(a1-a) < eps2 ) goto L50;}

  for(i=0; i<n; i++)
   {
    if(fabs(x[i]) > 1.) 
	   {if( fabs(sm[i]/x[i]) < eps1) ks++;}
    else
	   if(fabs(sm[i]) < eps1) ks++;
   }
 if(ks==n) goto L50;

 goto L4;


L60:
  for(i=0; i<n; i++)
    {
	if(x1[i]!=x[i])
	 {
	  for(j=0; j<n; j++)
		  x1[j]=x[j];
	  if(neud==1) goto L50;
	  neud=1;
	  goto L4;
	 }
    }

L50:

 delete sm;
 delete yy1;
 delete sag;
 delete df;
 delete dx;
 delete x1;
 delete m;
 for(i=0;i<n;i++)  delete h[i];
 delete h;
 return a;
}

/*

     **************************************************
        
         a
	   :
     a- n*n,
     k-  ,
     l-  ,
     **************************************************

*/
 void gord(double** a,int n,int k,int l)
{
 double q;
 int i,j;
  for(i=0; i<n; i++)
	for(j=0; j<n; j++)
      {
	 if(i!=k && j!=l)
	   a[i][j]=a[i][j]-a[k][j]*a[i][l]/a[k][l] ;
	}
 q=a[k][l] ;
  for(j=0; j<n; j++)
	  a[k][j]=-a[k][j]/q ;
  for(i=0; i<n; i++)
	a[i][l]=a[i][l]/q ;
  a[k][l]=-a[k][l] ;
}

/*
   **************************************************
         
	 :
     a1-p n*n.
	 :
     a-  n*n;
     iflag -  : 1 -  
	      , 0 -  a1 .
   **************************************************
*/

 void inv(double** a,int n,int *iflag)
{
 int i,j,k,kj,l;

 *iflag=TRUE;
 for(i=0; i<n; i++)	 m[i]=i ;

 k=n ;
 j=0 ;

 while( k > 0)
 {
 i=m[j] ;
 if(a[i][i]!=0.)
  {
    gord(a,n,i,i);
    kj=k-j-1 ;
    if(kj != 0)
    for(l=1; l<=kj; l++)
		   m[j+l-1]=m[j+l];

    k--;
    j=0;
  }
 else
  {
   j=j+1 ;
   if(j > k)
	{
	  *iflag=FALSE;
	  return;
	}
  }
 }
}

/*  p  */

void Hesse(double (*f)(double*), double x[], double** dx, int n)
{
 double h = 1e-8, h2 = 2*h, *x1, f1, f2, f3, f4;
 int i, j;

 x1 = new double[n];     
 
  for(i = 0; i < n; i++) x1[i] = x[i];

  for(i = 0; i < n; i++)
   for(j = 0; j < n; j++)
  {
   if(i != j)
  {
   x1[i] = x[i] + h;
   x1[j] = x[j] + h;
   f1 = f(x1);
   x1[i] = x[i] - h;
   f2 = f(x1);
   x1[j] = x[j] - h;
   f3 = f(x1);
   x1[i] = x[i] + h;
   f4 = f(x1);
   dx[i][j] = ((f1 - f2)/h2 + (f3 - f4)/h2)/h2;
  }
  else
  {
   x1[i] = x[i] + h;
   f1 = f(x1);
   x1[i] = x[i];
   f2 = f(x1);
   x1[i] = x[i] - h;
   f3 = f(x1);
   dx[i][i] = (f1 - 2*f2 + f3)/h/h;
  }
  x1[i] = x[i];
  x1[j] = x[j];
 }
 delete x1;
}