//(c) Artem Parlyuk, PhD., 
// e-mail: info@artsoft.name, http://artsoft.name
//(c) Boris Y. Lemeshko, Prof., 
// Department of Applied Mathematics and Computer Science
// Novosibirsk State Technical University, 
// e-mail: heard@fpm.ami.nstu.ru


#include "minimize.hpp"
#include <stdio.h>


/* p  */

int maxiteration=0, iteration, minerror, minimize_stop;
void (*Minimize_callback)(double, double x[], int) = NULL;


void SetMaxIteration(int x)
{
 maxiteration = x;
}

int GetMaxIteration()
{
 return maxiteration;
}


int GetIteration()
{
 return iteration;
}

int GetError()
{
 return minerror;
}

void StopMinimize()
{
 minimize_stop = 1;
}

void AllowMinimize()
{
 minimize_stop = 0;
}
             
void SetMinimizeCallBack(void (*callback)(double, double x[], int))
{
 Minimize_callback = callback;
}

double Restart(pMethod Method, double (*f)(double x[]),double x[],int n,double eps, double eps1)
{
  double res2 = Method(f,x,n,eps,eps1),res1;
  do
  {
   if(minimize_stop) break;
   res1 = res2;
   res2 = Method(f,x,n,eps,eps1);
  }
  while(res1 != res2);
  return res2;
}


 double* sigm;
 void blrev(double al[],double** d,int n,int *b);

 double Rosen(double (*f)(double x[]),double x[],int n,
              double eps , double eps1)
{
  //double  sag[10], al[10], dx[10], x1[10], xx[10], d[10][10];
  double  a1, r, t, s, t1, dla, la, a;
  int  i, j, k, kt, b;
  double  *sag, *al, *dx, *x1, *xx;
  double** d;

  sag = new double[n];
  al = new double[n];
  dx = new double[n];
  x1 = new double[n];
  xx = new double[n];
  d = new double*[n];
  sigm = new double[3*n];
  for(i=0;i<n;i++) d[i] = new double[n];

   iteration = 0;
   minerror = 0;


/*     */

 for(i=0; i<n; i++)
    {
	al[i]=0.1;
	x1[i]=x[i];
    }
 for(i=0; i<n; i++)
    for(j=0; j<n; j++)
       {
	 d[i][j]=0.;
	 if(i==j) d[i][j]=1.;
       }
    a=(*f)(x);

/*   */

L3:
/*      printf("f(x) = %lf\t(", a);
      for(j = 0; j < n; j++)
       printf("%lf ",x[j]);
      printf(")\n");
 */
        if(Minimize_callback) Minimize_callback(a, x, n);
	if((maxiteration > 0) && (iteration > maxiteration))
	{
	 minerror = ERR_ITER;
	 goto L20;
	}
        if(minimize_stop)
	{
	 minerror = ERR_ITER;
	 goto L20;
	}
	iteration++;
 a1=a;
    for(j=0; j<n; j++)
      {
	la=al[j];
	for(i=0; i<n; i++) sag[i]=d[i][j];
	if(la==0.) la=0.01;

/*   */

L101:   k=0;
	kt=0;
L103:   for(i=0; i<n; i++) xx[i]=x[i]+la*sag[i];
	r=(*f)(xx);
	if(r >= a) goto L107;
	if(k==1) goto L190;
L108:   for(i=0; i<n; i++) xx[i]=xx[i]+la*sag[i];
	t=(*f)(xx);
	if(t >= r) goto L190;
	la=la*2;
	r=t;
	goto L108;
L107:   for(i=0; i<n; i++) xx[i]=x[i]-la*sag[i];
	s=(*f)(xx);
	if(a <= s)
	{
	  la=la/2;
	  kt=kt+1;
	  if(15 <= kt) { la=0.; goto L26; }
	  k=1;
	  t=r;
	  t1=s;
	  goto L103;
	}
	la=-la;
	r=s;
	if(k != 1) goto L108;
	t=t1;
L190: /*   */
	s=a+t-2*r;
	if(s == 0.) { dla=la; goto L121; }
	dla=la*(3*a+t-4*r)/(s*2);
	if(dla == 0.) { la=0.; goto L26; }
L121:
  	for(i=0; i<n; i++) xx[i]=x[i]+dla*sag[i];
	a=(*f)(xx);
	if(a <= r)
	{
	 la=dla;
	 for(i=0; i<n; i++) x[i]=xx[i];
	}
	else
	{
	  a=r;
	  for(i=0; i<n; i++) x[i]=x[i]+la*sag[i];
	}
	al[j]=la;
L26: ;
  }
	for(i=0; i<n; i++)
	{
	  dx[i]=x[i]-x1[i];
	  x1[i]=x[i];
	}
	if(1. > fabs(a)) goto L7;
	if(fabs(a-a1) < eps1) goto L20;
	goto L55;
L7: 	if(fabs(a) < eps1) goto L20;
L55:  	for(i=0; i<n; i++)
	{
	   if(fabs(xx[i]) < 1.) goto L13;
	   if(fabs(dx[i]/x[i]) > eps) goto L12;
	   goto L9;
L13: 	  if(fabs(dx[i]) > eps) goto L12;
L9: ;
  }
	goto L20;
L12: 	blrev(al,d,n,&b);
	if(!b) goto L20;
	goto L3;
L20:  

  delete sag;
  delete al;
  delete dx;
  delete x1;
  delete xx;
  for(i=0;i<n;i++) delete d[i];
  delete d;
  delete sigm;

  return a;
}

 void blrev(double al[],double** d,int n,int *b)
/*    */
{
 double s, p;
 int nt, i, j;

     *b=TRUE;
     nt=n-1;
     i=n-1;
L50: if(al[i] != 0.) goto L2;
     al[i]=al[nt];
     al[nt]=0.;
     for(j=0; j<n; j++)
     {
	 sigm[j]=d[j][i];
	 d[j][i]=d[j][nt];
	 d[j][nt]=sigm[j];
     }
     nt=nt-1;
L2:  i=i-1;
     if(i >= 0) goto L50;
     if(nt == -1) { *b=FALSE; return; }
     s=0.;
     for(i=0; i<n; i++)  sigm[i]=0.;
L10: s=s+al[nt]*al[nt];
     for(i=0; i<n; i++)  sigm[i]=sigm[i]+al[nt]*d[i][nt];
     if(nt == 0) goto L15;
     p=sqrt(s*(s+al[nt-1]*al[nt-1]));
     if(p == 0.) goto L17;
     for(i=0; i<n; i++)  d[i][nt]=(s*d[i][nt-1]-al[nt-1]*sigm[i])/p;
L17: nt=nt-1;
     if(nt > -1) goto L10;
L15: p=sqrt(s);
     if(p == 0.) {*b=FALSE; return; }
     for(i=0; i<n; i++)  d[i][nt]=sigm[i]/p;
     return;
}

