unit minimizelib;

interface
uses windows;
const
 MINIMIZE_ERR_OK = 0;
 MINIMIZE_ERR_ITER = 1;
 DEF_PRECISION = 1.e-8;

type
 ppdouble = ^pdouble;
 pcallback = procedure (a : double; x : pdouble; n:integer); cdecl;
 pfunction = function (x : pdouble) : double; cdecl;
 pgrad =  procedure (fun : pfunction; x:pdouble; res:pdouble; n:integer); cdecl;
 phesse = procedure (fun : pfunction; x : pdouble; res : ppdouble; n : integer); cdecl;
 pMethod = function (fun : pfunction; x : pdouble; n: integer; eps1 : double; eps2 : double) : double; cdecl;
 pMethod2 = function (fun : pfunction; x: pdouble; n : integer;
               param : integer = 1; eps1 : double = DEF_PRECISION; eps2 : double = DEF_PRECISION) : double;  cdecl;
 pSetMinimizeCallBack = procedure (proc : pcallback);  cdecl; 
 pSetMaxIteration = procedure (x:integer);  cdecl;
 pGetfunction = function : integer;  cdecl; 
 pProc = procedure; cdecl; 
 pSetGrad = procedure (fun : pgrad);  cdecl; 
 pSetHesse = procedure (fun : phesse);  cdecl;
 pRestart = function (Method: pMethod; fun : pfunction; x : pdouble; n : integer;
                 eps1 : double = DEF_PRECISION;
                 eps2 : double = DEF_PRECISION): double; cdecl; 

var SetMaxIteration : pSetMaxIteration;
    GetMaxIteration : pGetfunction;
    GetIteration : pGetfunction;
    GetError : pGetfunction;
    StopMinimize : pProc;
    AllowMinimize : pProc;
    SetMinimizeCallBack : pSetMinimizeCallBack;
	  SetGrad : pSetGrad;
  	SetHesse : pSetHesse;

    Restart : pRestart;


//Rosenbrok method
var Rosen : pMethod;

//Quasi-Newton algorithms
   QuasiNewton : pMethod2;

//Davidon-Fletcher-Powell
  Davidon : pMethod;
//Fletcher
  Fletcher : pMethod;
//Broyden
  Broyden : pMethod;

//------------------------------------------------------------
//Nonlinear conjugate gradient algorithms
   CGrad : pMethod2;

//Fletcher-Reeves
  Reeves : pMethod;
//Pshenichny
  Pshenichny : pMethod;
//Method of steepest descent
  SDescent : pMethod;

//------------------------------------------------------------
//Newton algorithm

  Newton : pMethod;

//------------------------------------------------------------
//Numeric gradient & hessian
var Grad : pgrad;
    Hesse : phesse;

var DllHandle : cardinal;

function InitDLL(dllpath:string):Cardinal;
procedure DeInitDLL;

implementation
uses sysutils;

function InitDll(dllpath:string):Cardinal;
var fp:string;
begin
 if DllHandle <> 0 then
 begin
  result := DllHandle;
  exit;
 end;
 SetErrorMode( SEM_FAILCRITICALERRORS or SEM_NOGPFAULTERRORBOX );
 try
      if dllpath = '' then fp := 'minimizelib.dll'
      else
       if dllpath[length(dllpath)] = '\' then
        fp := dllpath+'minimizelib.dll'
       else
        fp := dllpath;
      result := LoadLibrary(pchar(fp));
      DLLHandle := result;
      if result > HInstance_Error then
      begin

        @SetMaxIteration  := GetProcAddress(DllHandle,'_SetMaxIteration');
         if (@SetMaxIteration = nil)  then
            raise Exception.Create('MinimizeLib DLL is invalid!');

        @GetMaxIteration := GetProcAddress(DllHandle,'_GetMaxIteration');
         if (@GetMaxIteration = nil)  then
            raise Exception.Create('MinimizeLib DLL is invalid!');

        @GetIteration := GetProcAddress(DllHandle,'_GetIteration');
         if (@GetIteration = nil)  then
            raise Exception.Create('MinimizeLib DLL is invalid!');

        @GetError := GetProcAddress(DllHandle,'_GetError');
         if (@GetError = nil)  then
            raise Exception.Create('MinimizeLib DLL is invalid!');

        @StopMinimize := GetProcAddress(DllHandle,'_StopMinimize');
         if (@StopMinimize = nil)  then
            raise Exception.Create('MinimizeLib DLL is invalid!');

        @AllowMinimize := GetProcAddress(DllHandle,'_AllowMinimize');
         if (@AllowMinimize = nil)  then
            raise Exception.Create('MinimizeLib DLL is invalid!');

        @SetMinimizeCallBack := GetProcAddress(DllHandle,'_SetMinimizeCallBack');
         if (@SetMinimizeCallBack = nil)  then
            raise Exception.Create('MinimizeLib DLL is invalid!');

	      @SetGrad := GetProcAddress(DllHandle,'_SetGrad');
         if (@SetGrad = nil)  then
            raise Exception.Create('MinimizeLib DLL is invalid!');

  	    @SetHesse := GetProcAddress(DllHandle,'_SetHesse');
         if (@SetHesse = nil)  then
            raise Exception.Create('MinimizeLib DLL is invalid!');

        @Restart := GetProcAddress(DllHandle,'_Restart');
         if (@Restart = nil)  then
            raise Exception.Create('MinimizeLib DLL is invalid!');

        @Rosen := GetProcAddress(DllHandle,'_Rosen');
         if (@Rosen = nil)  then
            raise Exception.Create('MinimizeLib DLL is invalid!');

        @QuasiNewton := GetProcAddress(DllHandle,'_QuasiNewton');
         if (@QuasiNewton = nil)  then
            raise Exception.Create('MinimizeLib DLL is invalid!');

        @Davidon := GetProcAddress(DllHandle,'_Davidon');
         if (@Davidon = nil)  then
            raise Exception.Create('MinimizeLib DLL is invalid!');

        @Fletcher := GetProcAddress(DllHandle,'_Fletcher');
         if (@Fletcher = nil)  then
            raise Exception.Create('MinimizeLib DLL is invalid!');

        @Broyden := GetProcAddress(DllHandle,'_Broyden');
         if (@Broyden = nil)  then
            raise Exception.Create('MinimizeLib DLL is invalid!');

        @CGrad := GetProcAddress(DllHandle,'_CGrad');
         if (@CGrad = nil)  then
            raise Exception.Create('MinimizeLib DLL is invalid!');

        @Reeves := GetProcAddress(DllHandle,'_Reeves');
         if (@Reeves = nil)  then
            raise Exception.Create('MinimizeLib DLL is invalid!');

        @Pshenichny := GetProcAddress(DllHandle,'_Pshenichny');
         if (@Pshenichny = nil)  then
            raise Exception.Create('MinimizeLib DLL is invalid!');

        @SDescent := GetProcAddress(DllHandle,'_SDescent');
         if (@SDescent = nil)  then
            raise Exception.Create('MinimizeLib DLL is invalid!');

        @Newton := GetProcAddress(DllHandle,'_Newton');
         if (@Newton = nil)  then
            raise Exception.Create('MinimizeLib DLL is invalid!');

        @Grad := GetProcAddress(DllHandle,'_Grad');
         if (@Grad = nil)  then
            raise Exception.Create('MinimizeLib DLL is invalid!');

        @Hesse := GetProcAddress(DllHandle,'_Hesse');
         if (@Hesse = nil)  then
            raise Exception.Create('MinimizeLib DLL is invalid!');
      end
      else
        raise Exception.Create('MinimizeLib DLL not found!')
  except
   on E:Exception do
   begin
    SetErrorMode(0);
    raise;
   end;
  end;
end;

procedure DeInitDLL;
begin
 if DllHAndle <> 0 then
 begin
  FreeLibrary(DllHandle);
  DllHAndle := 0;
 end; 
end;

end.
